/*
 * Copyright (c) 2014 F. Mannhardt (f.mannhardt@tue.nl)
 * 
 * LICENSE:
 * 
 * This code is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 3 of the License, or (at your option) any
 * later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA
 */

package org.processmining.xeslite.external;

import java.util.AbstractSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import org.deckfour.xes.model.XAttribute;
import org.deckfour.xes.model.XAttributeMap;

import com.google.common.base.Function;
import com.google.common.collect.Iterators;

/**
 * A {@link Set} implementation that provides a view on the entries of an
 * {@link XAttributeMap}. This implementation queries the underlying MapDB
 * storage and tries to avoid copies where possible. It implements the optional
 * {@link Entry#setValue(Object)} method of the {@link Entry} interface. It does
 * not allow for {@link MapDBEntrySet#add(Entry)}.
 * 
 * @author F. Mannhardt
 *
 */
final class MapDBEntrySet extends AbstractSet<Entry<String, XAttribute>> {

	private final static class MutableEntry implements Map.Entry<String, XAttribute> {

		private final MapDBAttributeMap attributeMap;

		private ExternalAttribute currentValue;

		public MutableEntry(ExternalAttribute attribute, MapDBAttributeMap attributeMap) {
			this.currentValue = attribute;
			this.attributeMap = attributeMap;
		}

		public String getKey() {
			return currentValue.getKey();
		}

		public XAttribute getValue() {
			return currentValue;
		}

		public XAttribute setValue(XAttribute value) {
			ExternalAttribute newAttribute = XAttributeExternalImpl.convert(attributeMap.getStore(), value);
			// Sanity check for invalid usage of this MutableEntry  
			if (currentValue.getInternalKey() != newAttribute.getInternalKey()) {
				throw new IllegalArgumentException(
						"Cannot change the 'key' of an attribute with the 'setValue' method! Trying to change "
								+ currentValue + " into " + newAttribute);
			}
			currentValue = newAttribute;
			return attributeMap.put(value.getKey(), newAttribute);
		}

		public boolean equals(Object obj) {
			if (this == obj) {
				return true;
			}
			if (!(obj instanceof Map.Entry)) {
				return false;
			}
			@SuppressWarnings("rawtypes")
			Map.Entry other = (Entry) obj;
			return getKey() == null ? other.getKey() == null
					: getKey().equals(other.getKey()) && getValue() == null ? other.getValue() == null : getValue()
							.equals(other.getValue());
		}

		public int hashCode() {
			return (getKey() == null ? 0 : getKey().hashCode()) ^ (getValue() == null ? 0 : getValue().hashCode());
		}
	}

	private final MapDBAttributeMap attributeMap;

	MapDBEntrySet(MapDBAttributeMap attributeMap) {
		this.attributeMap = attributeMap;
	}

	@Override
	public Iterator<Entry<String, XAttribute>> iterator() {
		// Iteration over the values() is enough, as the Attribute also contains its key
		return Iterators.transform(attributeMap.getInternalSubMap().entrySet().iterator(),
				new Function<Entry<Long, ExternalAttribute>, Entry<String, XAttribute>>() {

					@Override
					public Entry<String, XAttribute> apply(Entry<Long, ExternalAttribute> entry) {
						Long compositeKey = entry.getKey();
						int attributeKey = MapDBStore.getInternalAttributeKey(attributeMap.getOwner(),
								compositeKey);

						// Decorate before wrapping in Entry
						ExternalAttribute attribute = XAttributeExternalImpl.decorate(entry.getValue(),
								attributeKey, attributeMap.getStore());
						return new MutableEntry(attribute, attributeMap);
					}
				});
	}

	@Override
	public int size() {
		return attributeMap.size();
	}

	// Overridden for efficiency

	@Override
	public boolean contains(Object o) {
		String attributeKey = getAttributeKey(o);
		if (attributeKey != null) {
			XAttribute storedAttribute = attributeMap.get(attributeKey);
			if (storedAttribute != null) {
				XAttribute attribute = getAttribute(o);
				return attribute.equals(storedAttribute);
			} else {
				return false;
			}
		}
		return false;
	}

	@Override
	public boolean remove(Object o) {
		String attributeKey = getAttributeKey(o);
		if (attributeKey != null) {
			return attributeMap.remove(attributeKey) != null;
		}
		return false;
	}

	@Override
	public void clear() {
		attributeMap.clear();
	}

	private static XAttribute getAttribute(Object o) {
		if (o == null || (!(o instanceof Entry<?, ?>))) {
			return null; // null is not permitted
		}
		Entry<?, ?> entry = (Entry<?, ?>) o;
		Object value = entry.getValue();
		if (value == null || (!(value instanceof XAttribute))) {
			return null;
		}
		return (XAttribute) value;
	}

	private static String getAttributeKey(Object o) {
		if (o == null || (!(o instanceof Entry<?, ?>))) {
			return null; // null is not permitted
		}
		Entry<?, ?> entry = (Entry<?, ?>) o;
		Object key = entry.getKey();
		if (key == null || (!(key instanceof String))) {
			return null;
		}
		return (String) key;
	}

}