package org.processmining.xeslite.plugin;

import java.io.Closeable;
import java.io.File;
import java.io.FileInputStream;
import java.io.InputStream;
import java.security.InvalidParameterException;
import java.util.Collection;
import java.util.Enumeration;
import java.util.zip.GZIPInputStream;
import java.util.zip.ZipEntry;
import java.util.zip.ZipFile;

import org.deckfour.xes.extension.std.XConceptExtension;
import org.deckfour.xes.factory.XFactory;
import org.deckfour.xes.in.XMxmlParser;
import org.deckfour.xes.in.XParser;
import org.deckfour.xes.model.XLog;
import org.processmining.framework.abstractplugins.AbstractImportPlugin;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.plugins.log.XContextMonitoredInputStream;
import org.processmining.xeslite.external.XFactoryExternalStore;
import org.processmining.xeslite.parser.XesLiteXmlParser;

import com.google.common.base.Throwables;
import com.google.common.collect.Iterators;

public abstract class AbstractOpenLogFileLitePlugin extends AbstractImportPlugin {

	// Have to keep that here because the 'AbstractImportPlugin' does not care about closing
	//TODO close in original class
	private Closeable currentResource;

	abstract protected void beforeParse(XFactory factory);

	abstract protected void afterParse(XFactory factory);

	protected Object importFromUsingFactory(PluginContext context, InputStream input, String fileName,
			long fileSizeInBytes, XFactory factory) throws Exception {

		beforeParse(factory);
		Collection<XLog> logs = parseLog(fileName,
				new XContextMonitoredInputStream(input, fileSizeInBytes, context.getProgress()), factory);
		afterParse(factory);		

		if (logs.size() == 0) {
			throw new Exception("No processes contained in log! Parser could did not find any log.");
		}

		XLog log = Iterators.getOnlyElement(logs.iterator());
		if (XConceptExtension.instance().extractName(log) == null) {
			XConceptExtension.instance().assignName(log, "Anonymous log imported from " + fileName);
		}

		if (context != null) {
			context.getFutureResult(0).setLabel(XConceptExtension.instance().extractName(log));
		}
		
		return log;
	}

	private Collection<XLog> parseLog(String fileName, XContextMonitoredInputStream input, XFactory factory)
			throws Exception {

		XParser parser;
		if (isXES(fileName)) {
			parser = new XesLiteXmlParser(factory, true);
		} else {
			if (!(factory instanceof XFactoryExternalStore)) {
				parser = new XMxmlParser(factory);
			} else {
				throw new Exception("This importer does not support parsing MXML! Please use a different importer!");
			}
		}

		try {
			return parser.parse(input);
		} catch (Exception e) {
			throw new Exception(Throwables.getRootCause(e).getMessage() + "\n\nStacktrace for reporting:\n"
					+ Throwables.getStackTraceAsString(e));
		} finally {
			// Clean-up
			if (currentResource != null) {
				currentResource.close();
			}
		}

	}

	protected boolean isXES(String filename) {
		return filename.toLowerCase().endsWith(".xes") || filename.toLowerCase().endsWith(".xez")
				|| filename.toLowerCase().endsWith(".xes.gz")
				|| (filename.toLowerCase().endsWith("zip") && !filename.toLowerCase().contains("mxml")); // just assume it is a XES file in there
	}

	@Override
	protected InputStream getInputStream(File file) throws Exception {
		FileInputStream stream = new FileInputStream(file);
		currentResource = stream;
		if (file.getName().endsWith(".gz") || file.getName().endsWith(".xez")) {
			GZIPInputStream inputStream = new GZIPInputStream(stream);
			currentResource = inputStream;
			return inputStream;
		}
		if (file.getName().endsWith(".zip")) {
			ZipFile zip = new ZipFile(file);
			currentResource = zip;
			Enumeration<? extends ZipEntry> entries = zip.entries();
			ZipEntry zipEntry = entries.nextElement();
			if (entries.hasMoreElements()) {
				throw new InvalidParameterException("Zipped log files should not contain more than one entry.");
			}
			return zip.getInputStream(zipEntry);
		}		
		return stream;
	}

}
